@@ -70,6 +70,7 @@ enum Error {
     NoArtifacts,
     UnsupportedFileFormat(path::PathBuf),
     ParsingError(binfarce::ParseError),
+    PdbError(pdb::Error),
 }
 
 impl From<binfarce::ParseError> for Error {
@@ -84,6 +85,12 @@ impl From<binfarce::UnexpectedEof> for Error {
     }
 }
 
+impl From<pdb::Error> for Error {
+    fn from(e: pdb::Error) -> Self {
+        Error::PdbError(e)
+    }
+}
+
 impl fmt::Display for Error {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match *self {
@@ -121,6 +128,9 @@ impl fmt::Display for Error {
             Error::ParsingError(ref e) => {
                 write!(f, "parsing failed cause '{}'", e)
             }
+            Error::PdbError(ref e) => {
+                write!(f, "error parsing pdb file cause '{}'", e)
+            }
         }
     }
 }
@@ -262,10 +272,12 @@ OPTIONS:
         --features <FEATURES>       Space-separated list of features to activate
         --all-features              Activate all available features
         --no-default-features       Do not activate the `default` feature
+        --profile <PROFILE>         Build with the given profile.
         --target <TARGET>           Build for the target triple
         --target-dir <DIRECTORY>    Directory for all generated artifacts
         --frozen                    Require Cargo.lock and cache are up to date
         --locked                    Require Cargo.lock is up to date
+    -Z <FLAG>...                    Unstable (nightly-only) flags to Cargo, see 'cargo -Z help' for details
         --crates                    Per crate bloatedness
         --time                      Per crate build time. Will run `cargo clean` first
         --filter <CRATE|REGEXP>     Filter functions by crate
@@ -306,10 +318,12 @@ pub struct Args {
     features: Option<String>,
     all_features: bool,
     no_default_features: bool,
+    profile: Option<String>,
     target: Option<String>,
     target_dir: Option<String>,
     frozen: bool,
     locked: bool,
+    unstable: Vec<String>,
     crates: bool,
     time: bool,
     filter: Option<String>,
@@ -339,10 +353,12 @@ fn parse_args(raw_args: Vec<std::ffi::OsString>) -> Result<Args, pico_args::Erro
         features:               input.opt_value_from_str("--features")?,
         all_features:           input.contains("--all-features"),
         no_default_features:    input.contains("--no-default-features"),
+        profile:                input.opt_value_from_str("--profile")?,
         target:                 input.opt_value_from_str("--target")?,
         target_dir:             input.opt_value_from_str("--target-dir")?,
         frozen:                 input.contains("--frozen"),
         locked:                 input.contains("--locked"),
+        unstable:               input.values_from_str("-Z")?,
         crates:                 input.contains("--crates"),
         time:                   input.contains("--time"),
         filter:                 input.opt_value_from_str("--filter")?,
@@ -507,35 +523,52 @@ fn get_workspace_root() -> Result<String, Error> {
 fn process_crate(args: &Args) -> Result<CrateData, Error> {
     let workspace_root = get_workspace_root()?;
 
+    let default_target = get_default_target()?;
+    let target_triple = args.target.clone().unwrap_or_else(|| default_target);
+
     let child = if args.time {
         // To collect the build times we have to clean the repo first.
-
-        let clean_args = if args.release {
-            // Remove only `target/release` in the release mode.
-            vec!["clean", "--release"]
-        } else {
-            // We can't remove only `target/debug` in debug mode yet.
-            // See https://github.com/rust-lang/cargo/pull/6989
-            vec!["clean"]
-        };
-
-        // No need to check the output status.
-        let _ = Command::new("cargo")
-            .args(&clean_args)
-            .output();
+        cargo_clean(args.release);
 
         Command::new("cargo")
-            .args(&get_cargo_args(args))
+            .args(&get_cargo_args(args, true))
             .env("RUSTC_WRAPPER", "cargo-bloat")
             .stdout(std::process::Stdio::piped())
             // Hide cargo output, because we are using stderr to track build time.
             .stderr(std::process::Stdio::piped())
             .spawn().map_err(|_| Error::CargoBuildFailed)?
     } else {
-        Command::new("cargo")
-            .args(&get_cargo_args(args))
-            .stdout(std::process::Stdio::piped())
-            .spawn().map_err(|_| Error::CargoBuildFailed)?
+        // Run `cargo build` without json output first, so we could print build errors.
+        {
+            let cmd = &mut Command::new("cargo");
+            cmd.args(&get_cargo_args(args, false));
+
+            // When targeting MSVC, symbols data will be stored in PDB files.
+            // But unlike other targets, the Release build would not have any useful information.
+            // Therefore we have to force debug info in Release mode for MSVC target.
+            if args.release && target_triple.contains("msvc") {
+                cmd.env("CARGO_PROFILE_RELEASE_DEBUG", "true");
+            }
+
+            cmd.spawn().map_err(|_| Error::CargoBuildFailed)?
+                .wait().map_err(|_| Error::CargoBuildFailed)?;
+        }
+
+        // Run `cargo build` with json output and collect it.
+        // This would not cause a rebuild.
+        let cmd = &mut Command::new("cargo");
+        cmd.args(&get_cargo_args(args, true));
+        cmd.stdout(std::process::Stdio::piped());
+        cmd.stderr(std::process::Stdio::null());
+
+        // When targeting MSVC, symbols data will be stored in PDB files.
+        // But unlike other targets, the Release build would not have any useful information.
+        // Therefore we have to force debug info in Release mode for MSVC target.
+        if args.release && target_triple.contains("msvc") {
+            cmd.env("CARGO_PROFILE_RELEASE_DEBUG", "true");
+        }
+
+        cmd.spawn().map_err(|_| Error::CargoBuildFailed)?
     };
 
     let output = child.wait_with_output().map_err(|_| Error::CargoBuildFailed)?;
@@ -610,6 +643,10 @@ fn process_crate(args: &Args) -> Result<CrateData, Error> {
     }
 
     if args.time && !args.crates {
+        // Clean target again, because cargo will cache RUSTC_WRAPPER,
+        // which is not what we want.
+        cargo_clean(args.release);
+
         // We don't care about symbols if we only plan to print the build times.
         return Ok(CrateData {
             exe_path: None,
@@ -621,9 +658,6 @@ fn process_crate(args: &Args) -> Result<CrateData, Error> {
         });
     }
 
-    let default_target = get_default_target()?;
-    let target_triple = args.target.clone().unwrap_or_else(|| default_target);
-
     let target_dylib_path = stdlibs_dir(&target_triple)?;
 
     let mut rlib_paths = Vec::new();
@@ -678,10 +712,13 @@ fn process_crate(args: &Args) -> Result<CrateData, Error> {
     Err(Error::UnsupportedCrateType)
 }
 
-fn get_cargo_args(args: &Args) -> Vec<String> {
+fn get_cargo_args(args: &Args, json_output: bool) -> Vec<String> {
     let mut list = Vec::new();
     list.push("build".to_string());
-    list.push("--message-format=json".to_string());
+
+    if json_output {
+        list.push("--message-format=json".to_string());
+    }
 
     if args.release {
         list.push("--release".to_string());
@@ -722,6 +759,11 @@ fn get_cargo_args(args: &Args) -> Vec<String> {
         list.push("-v".into());
     }
 
+    if let Some(ref profile) = args.profile {
+        list.push(format!("--profile={}", profile));
+    }
+
+
     if let Some(ref target) = args.target {
         list.push(format!("--target={}", target));
     }
@@ -738,6 +780,10 @@ fn get_cargo_args(args: &Args) -> Vec<String> {
         list.push("--locked".to_string());
     }
 
+    for arg in &args.unstable {
+        list.push(format!("-Z={}", arg));
+    }
+
     if let Some(jobs) = args.jobs {
         list.push(format!("-j{}", jobs));
     }
@@ -745,6 +791,22 @@ fn get_cargo_args(args: &Args) -> Vec<String> {
     list
 }
 
+fn cargo_clean(release: bool) {
+    let clean_args = if release {
+        // Remove only `target/release` in the release mode.
+        vec!["clean", "--release"]
+    } else {
+        // We can't remove only `target/debug` in debug mode yet.
+        // See https://github.com/rust-lang/cargo/pull/6989
+        vec!["clean"]
+    };
+
+    // No need to check the output status.
+    let _ = Command::new("cargo")
+        .args(&clean_args)
+        .output();
+}
+
 fn collect_rlib_paths(deps_dir: &path::Path) -> Vec<(String, path::PathBuf)> {
     let mut rlib_paths: Vec<(String, path::PathBuf)> = Vec::new();
     if let Ok(entries) = fs::read_dir(deps_dir) {
@@ -855,15 +917,135 @@ fn collect_macho_data(data: &[u8]) -> Result<Data, Error> {
     Ok(d)
 }
 
-fn collect_pe_data(path: &path::Path, data: &[u8]) -> Result<Data, Error> {
-    let (symbols, text_size) = pe::parse(data)?.symbols()?;
+fn collect_pdb_data(pdb_path: &path::Path, text_size: u64) -> Result<Data, Error> {
+    use pdb::FallibleIterator;
 
-    // `pe::parse` will return zero symbols for an executable built with MSVC.
-    if symbols.is_empty() {
-        eprintln!("Warning: MSVC target is not supported.");
-        return Err(Error::UnsupportedFileFormat(path.to_owned()))
+    let file = fs::File::open(&pdb_path).map_err(|_| Error::OpenFailed(pdb_path.to_owned()))?;
+    let mut pdb = pdb::PDB::open(file)?;
+
+    let dbi = pdb.debug_information()?;
+    let symbol_table = pdb.global_symbols()?;
+    let address_map = pdb.address_map()?;
+
+    let mut out_symbols = Vec::new();
+
+    // Collect the PublicSymbols.
+    let mut public_symbols = Vec::new();
+
+    let mut symbols = symbol_table.iter();
+    while let Ok(Some(symbol)) = symbols.next() {
+        match symbol.parse() {
+            Ok(pdb::SymbolData::Public(data)) => {
+                if data.code || data.function {
+                    public_symbols.push((data.offset, data.name.to_string().into_owned()));
+                }
+            }
+            _ => {}
+        }
     }
 
+    let mut modules = dbi.modules()?;
+    while let Some(module) = modules.next()? {
+        let info = match pdb.module_info(&module)? {
+            Some(info) => info,
+            None => continue,
+        };
+        let mut symbols = info.symbols()?;
+        while let Some(symbol) = symbols.next()? {
+            if let Ok(pdb::SymbolData::Public(data)) = symbol.parse() {
+                if data.code || data.function {
+                    public_symbols.push((data.offset, data.name.to_string().into_owned()));
+                }
+            }
+        }
+    }
+
+    let cmp_offsets = |a: &pdb::PdbInternalSectionOffset, b: &pdb::PdbInternalSectionOffset| {
+        a.section.cmp(&b.section).then(a.offset.cmp(&b.offset))
+    };
+    public_symbols.sort_unstable_by(|a, b| cmp_offsets(&a.0, &b.0));
+
+    // Now find the Procedure symbols in all modules
+    // and if possible the matching PublicSymbol record with the mangled name.
+    let mut handle_proc = |proc: pdb::ProcedureSymbol| {
+        let mangled_symbol = public_symbols
+            .binary_search_by(|probe| {
+                let low = cmp_offsets(&probe.0, &proc.offset);
+                let high = cmp_offsets(&probe.0, &(proc.offset + proc.len));
+
+                use std::cmp::Ordering::*;
+                match (low, high) {
+                    // Less than the low bound -> less.
+                    (Less, _) => Less,
+                    // More than the high bound -> greater.
+                    (_, Greater) => Greater,
+                    _ => Equal,
+                }
+            })
+            .ok()
+            .map(|x| &public_symbols[x]);
+
+        let demangled_name = proc.name.to_string().into_owned();
+        out_symbols.push((
+            proc.offset.to_rva(&address_map),
+            proc.len as u64,
+            demangled_name,
+            mangled_symbol,
+        ));
+    };
+
+    let mut symbols = symbol_table.iter();
+    while let Ok(Some(symbol)) = symbols.next() {
+        if let Ok(pdb::SymbolData::Procedure(proc)) = symbol.parse() {
+            handle_proc(proc);
+        }
+    }
+
+    let mut modules = dbi.modules()?;
+    while let Some(module) = modules.next()? {
+        let info = match pdb.module_info(&module)? {
+            Some(info) => info,
+            None => continue,
+        };
+
+        let mut symbols = info.symbols()?;
+
+        while let Some(symbol) = symbols.next()? {
+            if let Ok(pdb::SymbolData::Procedure(proc)) = symbol.parse() {
+                handle_proc(proc);
+            }
+        }
+    }
+
+    let symbols = out_symbols
+        .into_iter()
+        .filter_map(|(address, size, unmangled_name, mangled_name)| {
+            if let Some(address) = address {
+                Some(SymbolData {
+                    name: mangled_name
+                        .map(|(_, mangled_name)| {
+                            binfarce::demangle::SymbolName::demangle(mangled_name)
+                        })
+                        // Assume the Symbol record name is unmangled if we didn't find one.
+                        // Note that unmangled names stored in PDB have a different format from
+                        // one stored in binaries itself. Specifically they do not include hash
+                        // and can have a bit different formatting.
+                        // We also assume that a Legacy mangling scheme were used.
+                        .unwrap_or_else(|| binfarce::demangle::SymbolName {
+                            complete: unmangled_name.clone(),
+                            trimmed: unmangled_name.clone(),
+                            crate_name: None,
+                            kind: binfarce::demangle::Kind::Legacy,
+                        }),
+                    address: address.0 as u64,
+                    size,
+                })
+            } else {
+                None
+            }
+        })
+        .collect();
+
     let d = Data {
         symbols,
         file_size: 0,
@@ -873,6 +1055,34 @@ fn collect_pe_data(path: &path::Path, data: &[u8]) -> Result<Data, Error> {
     Ok(d)
 }
 
+fn collect_pe_data(path: &path::Path, data: &[u8]) -> Result<Data, Error> {
+    let (symbols, text_size) = pe::parse(data)?.symbols()?;
+
+    // `pe::parse` will return zero symbols for an executable built with MSVC.
+    if symbols.is_empty() {
+        let pdb_path = {
+            let file_name = if let Some(file_name) = path.file_name() {
+                if let Some(file_name) = file_name.to_str() {
+                    file_name.replace("-", "_")
+                } else {
+                    return Err(Error::OpenFailed(path.to_owned()));
+                }
+            } else {
+                return Err(Error::OpenFailed(path.to_owned()));
+            };
+            path.with_file_name(file_name).with_extension("pdb")
+        };
+
+        collect_pdb_data(&pdb_path, text_size)
+    } else {
+        Ok(Data {
+            symbols,
+            file_size: 0,
+            text_size,
+        })
+    }
+}
+
 struct Methods {
     has_filter: bool,
     filter_out_size: u64,
